/*                                                           */
/* CDTrayEx.c by Andrew I. Reshin. Last modified 19.05.2001. */
/*                                                           */

#include <windows.h>
#include <winioctl.h>

#define MAIN_ICON		1
#define CDTRAY_DOS		"cdtray.exe"

//
// Show error message
//

ULONG Err(PCHAR Msg)
{
ULONG code,len;
PVOID errbuf;
PCHAR msgbuf;

code=GetLastError();
if(!Msg) Msg="";
errbuf=NULL;

len=
 FormatMessage(
	FORMAT_MESSAGE_IGNORE_INSERTS|
	FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
	0,code,MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),(PCHAR)&errbuf,0,0
	);

if(!len) len=sizeof("Unknown error")-1;

if(msgbuf=(PCHAR)LocalAlloc(LMEM_FIXED,lstrlen(Msg)+16+len))
 {
 wsprintf(msgbuf,"%s\n[#%u] %s",Msg,code,errbuf?(PCHAR)errbuf:"Unknown error");
 MessageBox(0,msgbuf,0,MB_OK);
 LocalFree(msgbuf);
 }

if(errbuf) LocalFree(errbuf);
return code;
}

//
// Call DOS programm CDTRAY.EXE for ejecting/closing CD-tray
//

BOOL CDTrayMscdex(CHAR Drive,CHAR Action)
{
CHAR buf[sizeof(CDTRAY_DOS)+sizeof(' ')+sizeof(Drive)+sizeof(Action)];
STARTUPINFO si={sizeof(STARTUPINFO)};
PROCESS_INFORMATION pi;
ULONG res;
BOOL isok;

wsprintf(buf,"%s %c%c",CDTRAY_DOS,Drive,Action);
si.dwFlags=STARTF_USESHOWWINDOW;
si.wShowWindow=SW_HIDE;

if(!CreateProcess(NULL,buf,NULL,NULL,FALSE,0,NULL,NULL,&si,&pi))
 return Err(CDTRAY_DOS);

do res=WaitForSingleObject(pi.hProcess,5000);
 while(
  res==WAIT_TIMEOUT &&
  MessageBox(0,"Wait next 5 seconds?","Time out",MB_OKCANCEL)==IDOK
  );

if(isok=(res==WAIT_OBJECT_0))
 {
 isok=GetExitCodeProcess(pi.hProcess,&res) && !res;
 if(!isok) MessageBox(NULL,CDTRAY_DOS,"Operation failed",MB_OK);
 }
else if(res!=WAIT_TIMEOUT) Err(CDTRAY_DOS);

CloseHandle(pi.hProcess);
CloseHandle(pi.hThread);
return isok;
}

//
// Eject/close CD-tray under WinNT or call CDTrayMscdex under Win9X
//

ULONG CDTray(CHAR Drive)
{
OSVERSIONINFO ovi;

ovi.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
if(!GetVersionEx(&ovi)) return Err("GetVersionEx");

if(ovi.dwPlatformId==VER_PLATFORM_WIN32_NT)
 {
 CHAR buf[]="\\\\.\\A:";
 HANDLE hdrive;
 ULONG dummy;

 buf[4]=Drive;

 hdrive=
  CreateFile(
   buf,GENERIC_READ,FILE_SHARE_WRITE|FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL
   );

 if(hdrive==INVALID_HANDLE_VALUE) return Err(buf);

 if(
  !DeviceIoControl(hdrive,IOCTL_DISK_EJECT_MEDIA,0,0,0,0,&dummy,0) || (
  MessageBox(0,"Would you like to close the tray?"," ",MB_OKCANCEL)==IDOK &&
  !DeviceIoControl(hdrive,IOCTL_DISK_LOAD_MEDIA,0,0,0,0,&dummy,0)
  ))
 Err(buf);

 CloseHandle(hdrive);
 }
else
 if(
  CDTrayMscdex(Drive,'O') &&
  MessageBox(0,"Would you like to close the tray?"," ",MB_OKCANCEL)==IDOK
  )
 CDTrayMscdex(Drive,'C');

return 0;
}

//
// Build and show menu with separate items for each CD-drive
//

void ShowMenu(HWND hWnd)
{
CHAR root[]="A:\\";
ULONG drives;
HMENU hmenu;
POINT pt;

hmenu=CreatePopupMenu();
InsertMenu(hmenu,0,MF_BYCOMMAND,0,"Exit");
InsertMenu(hmenu,0,MF_BYCOMMAND|MF_SEPARATOR,0,NULL);
drives=GetLogicalDrives();

for(*root='A'; *root<='Z'; (*root)++)
 if((drives & (1 << (*root-'A'))) && GetDriveType(root)==DRIVE_CDROM)
  InsertMenu(hmenu,0,MF_BYCOMMAND,*root,root);

SetForegroundWindow(hWnd);
GetCursorPos(&pt);
TrackPopupMenu(hmenu,0,pt.x,pt.y,0,hWnd,NULL);
PostMessage(hWnd,WM_NULL,0,0);
DestroyMenu(hmenu);
}

//
// Hidden window procedure. Icon/menu's messages handling.
//

LRESULT CALLBACK MainProc(HWND hWnd,UINT Message,WPARAM wParam,LPARAM lParam)
{
switch(Message)
 {
 case WM_USER:
   if(lParam==WM_RBUTTONDOWN || lParam==WM_LBUTTONDOWN) ShowMenu(hWnd);
   break;

 case WM_COMMAND:
   if(!lParam)
    if(LOWORD(wParam)>='A' && LOWORD(wParam)<='Z')
     CDTray((CHAR)LOWORD(wParam));
    else DestroyWindow(hWnd);
   break;

 case WM_DESTROY:
   PostQuitMessage(0);
   break;
 }

return DefWindowProc(hWnd,Message,wParam,lParam);
}

//
// Entry point. Create hidden window and add icon to the system tray.
//

int EntryPoint()
{
NOTIFYICONDATA nid;
WNDCLASS wc={0};
BOOL res;
MSG msg;

wc.lpfnWndProc=MainProc;
wc.hInstance=GetModuleHandle(NULL);
wc.lpszClassName="CDTrayEx";
RegisterClass(&wc);

nid.hWnd=
 CreateWindow("CDTrayEx","",WS_POPUP,0,0,0,0,NULL,NULL,wc.hInstance,NULL);

if(!(nid.hWnd)) return Err("CreateWindow");
nid.cbSize=sizeof(NOTIFYICONDATA);
nid.uID=0;
wsprintf(nid.szTip,"CDTrayEx");
nid.hIcon=LoadIcon(wc.hInstance,MAKEINTRESOURCE(MAIN_ICON));
nid.uCallbackMessage=WM_USER;
nid.uFlags=NIF_ICON|NIF_TIP|NIF_MESSAGE;
Shell_NotifyIcon(NIM_ADD,&nid);

while((res=GetMessage(&msg,NULL,0,0)) && res!=-1)
 {
 TranslateMessage(&msg);
 DispatchMessage(&msg);
 }

Shell_NotifyIcon(NIM_DELETE,&nid);
return 0;
}
